#
# Copyright (c) 2018, Salesforce, Inc.
#                     The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import List

import torch
import torch.nn as nn
from torch.jit import Final
from torch.nn import functional as F

INF = 1e10
EPSILON = 1e-10


class MultiLSTMCell(nn.Module):
    def __init__(self, num_layers, input_size, rnn_size, dropout):
        super(MultiLSTMCell, self).__init__()
        self.dropout = nn.Dropout(dropout)
        self.num_layers = num_layers
        self.layers = nn.ModuleList()

        for i in range(num_layers):
            self.layers.append(nn.LSTMCell(input_size, rnn_size))
            input_size = rnn_size

    def forward(self, input, hidden):
        h_0, c_0 = hidden
        h_1, c_1 = [], []
        for i, layer in enumerate(self.layers):
            input = self.dropout(input)
            h_1_i, c_1_i = layer(input, (h_0[i], c_0[i]))
            input = h_1_i
            h_1 += [h_1_i]
            c_1 += [c_1_i]

        h_1 = torch.stack(h_1)
        c_1 = torch.stack(c_1)

        return input, (h_1, c_1)


# torch.matmul can't do (4, 3, 2) @ (4, 2) -> (4, 3)
def matmul(x, y):
    if x.dim() == y.dim():
        return x @ y
    if x.dim() == y.dim() - 1:
        return (x.unsqueeze(-2) @ y).squeeze(-2)
    return (x @ y.unsqueeze(-2)).squeeze(-2)


class LayerNorm(nn.Module):
    def __init__(self, d_model, eps=1e-6):
        super().__init__()
        self.gamma = nn.Parameter(torch.ones(d_model))
        self.beta = nn.Parameter(torch.zeros(d_model))
        self.eps = eps

    def forward(self, x):
        mean = x.mean(-1, keepdim=True)
        std = x.std(-1, keepdim=True)
        return self.gamma * (x - mean) / (std + self.eps) + self.beta


def mask(targets, out, squash=True, pad_idx=1):
    mask = targets != pad_idx
    out_mask = mask.unsqueeze(-1).expand_as(out).contiguous()
    if squash:
        out_after = out[out_mask].contiguous().view(-1, out.size(-1))
    else:
        out_after = out * out_mask.float()
    targets_after = targets[mask]
    return out_after, targets_after


class LinearFeedforward(nn.Module):
    def __init__(self, d_in, d_hid, d_out, activation='relu', dropout=0.2):
        super().__init__()
        self.feedforward = Feedforward(d_in, d_hid, activation=activation)
        self.linear = Linear(d_hid, d_out)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        return self.dropout(self.linear(self.feedforward(x)))


class Linear(nn.Linear):
    def forward(self, x):
        size = x.size()
        return super().forward(x.contiguous().view(-1, size[-1])).view(*size[:-1], -1)


class Feedforward(nn.Module):
    def __init__(self, d_in, d_out, activation=None, bias=True, dropout=0.2):
        super().__init__()
        if activation is not None:
            self.activation = getattr(torch, activation)
        else:
            self.activation = lambda x: x
        self.linear = Linear(d_in, d_out, bias=bias)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        return self.activation(self.linear(self.dropout(x)))


class CombinedEmbedding(nn.Module):
    project: Final[bool]
    dimension: Final[int]

    def __init__(self, numericalizer, pretrained_embeddings, output_dimension, trained_dimension=0, project=True):
        super().__init__()
        self.project = project
        self.pretrained_embeddings = nn.ModuleList(pretrained_embeddings)

        dimension = 0
        for embedding in self.pretrained_embeddings:
            dimension += embedding.dim

        if trained_dimension > 0:
            self.trained_embeddings = nn.Embedding(numericalizer.num_tokens, trained_dimension)
            dimension += trained_dimension
        else:
            self.trained_embeddings = None

        if self.project:
            self.projection = Feedforward(dimension, output_dimension)
        else:
            assert dimension == output_dimension, (dimension, output_dimension)
        self.dimension = output_dimension

    def resize_embedding(self, new_vocab_size):
        if self.trained_embeddings is None:
            # we are not training embeddings at all
            return
        dimensions = self.trained_embeddings.weight.shape
        if new_vocab_size == dimensions[0]:
            return
        assert new_vocab_size > dimensions[0], 'Cannot shrink the embedding matrix'
        resized_embeddings = nn.Embedding(new_vocab_size, dimensions[1])
        resized_embeddings.weight.data[0 : dimensions[0], :] = self.trained_embeddings.weight.data
        self.trained_embeddings = resized_embeddings

    def _combine_embeddings(self, embeddings):

        emb = torch.cat(embeddings, dim=2)
        if self.project:
            emb = self.projection(emb)
        return emb

    def forward(self, x, padding=None):
        embedded: List[torch.Tensor] = []
        if self.pretrained_embeddings is not None:
            embedded += [emb(x, padding=padding) for emb in self.pretrained_embeddings]

        if self.trained_embeddings is not None:
            trained_vocabulary_size = self.trained_embeddings.weight.size()[0]
            valid_x = torch.lt(x, trained_vocabulary_size)
            masked_x = torch.where(valid_x, x, torch.zeros_like(x))
            output = self.trained_embeddings(masked_x)
            embedded.append(output)

        return self._combine_embeddings(embedded)


class LSTMDecoderAttention(nn.Module):
    def __init__(self, dim, dot=False):
        super().__init__()
        self.linear_in = nn.Linear(dim, dim, bias=False)
        self.linear_out = nn.Linear(2 * dim, dim, bias=False)
        self.tanh = nn.Tanh()
        self.mask = None
        self.dot = dot

    def applyMasks(self, context_mask):
        # context_mask is batch x encoder_time, convert it to batch x 1 x encoder_time
        self.context_mask = context_mask.unsqueeze(1)

    def forward(self, input: torch.Tensor, context: torch.Tensor):
        # input is batch x decoder_time x dim
        # context is batch x encoder_time x dim
        # output will be batch x decoder_time x dim
        # context_attention will be batch x decoder_time x encoder_time

        if not self.dot:
            targetT = self.linear_in(input)  # batch x decoder_time x dim x 1
        else:
            targetT = input

        transposed_context = torch.transpose(context, 2, 1)
        context_scores = torch.matmul(targetT, transposed_context)
        context_scores.masked_fill_(self.context_mask, -float('inf'))
        context_attention = F.softmax(context_scores, dim=-1) + EPSILON

        # convert context_attention to batch x decoder_time x 1 x encoder_time
        # convert context to batch x 1 x encoder_time x dim
        # context_alignment will be batch x decoder_time x 1 x dim
        context_alignment = torch.matmul(context_attention.unsqueeze(2), context.unsqueeze(1))
        # squeeze out the extra dimension
        context_alignment = context_alignment.squeeze(2)

        combined_representation = torch.cat([input, context_alignment], 2)
        output = self.tanh(self.linear_out(combined_representation))

        return output, context_attention


class LabelSmoothingCrossEntropy(torch.nn.Module):
    def __init__(self, smoothing):
        super(LabelSmoothingCrossEntropy, self).__init__()
        self.smoothing = smoothing

    def forward(self, x, target, ignore_index):
        """
        Inputs:
            x: Tensor of shape (N, vocab_size)
            target: Tensor of shape (N, ) where N is batch_size * sequence_length
            ignore_index: this index in the vocabulary is ignored when calculating loss. This is useful for pad tokens.
        Outputs:
            loss: a Tensor of shape (N, )
        """
        logprobs = F.log_softmax(x, dim=-1)
        nll_loss = -logprobs.gather(dim=-1, index=target.unsqueeze(1))
        nll_loss = nll_loss.squeeze(1)
        smooth_loss = -logprobs.mean(dim=-1)
        loss = (1.0 - self.smoothing) * nll_loss + self.smoothing * smooth_loss
        loss.masked_fill_((target == ignore_index), 0)
        return loss



#
# Copyright (c) 2019, The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
import math
import multiprocessing as mp
import os

import torch

from genienlp.data_utils.almond_utils import chunk_file, create_examples_from_file

from .base_dataset import Split
from .generic_dataset import CQA

logger = logging.getLogger(__name__)


class AlmondDataset(CQA):
    """Obtaining dataset for Almond semantic parsing task"""

    base_url = None

    def __init__(self, path, *, make_example, **kwargs):

        # TODO fix cache_path for multilingual task
        subsample = kwargs.get('subsample')
        cached_path = kwargs.get('cached_path')

        skip_cache = kwargs.get('skip_cache', True)
        cache_input_data = kwargs.get('cache_input_data', False)
        num_workers = kwargs.get('num_workers', 0)

        cache_name = os.path.join(cached_path, os.path.basename(path), str(subsample))
        dir_name = os.path.basename(os.path.dirname(path))

        if os.path.exists(cache_name) and not skip_cache:
            logger.info(f'Loading cached data from {cache_name}')
            examples = torch.load(cache_name)
        else:
            n = 0
            with open(path, 'r', encoding='utf-8') as fp:
                for line in fp:
                    n += 1

            max_examples = min(n, subsample) if subsample is not None else n
            if num_workers > 0:
                num_processes = min(num_workers, int(mp.cpu_count()))
                logger.info(f'Using {num_processes} workers...')
                chunk_size = int(math.ceil(max_examples / num_processes))
                num_chunks = int(math.ceil(max_examples / chunk_size))

                base_path, extension = path.rsplit('.', 1)

                chunk_file_paths = [f'{base_path}_{chunk_id}.tsv' for chunk_id in range(num_chunks)]
                chunk_file(path, chunk_file_paths, chunk_size, num_chunks)
                num_processes = min(num_processes, num_chunks)

                with mp.Pool(processes=num_processes) as pool:
                    process_args = [
                        {
                            'in_file': chunk_file_paths[i],
                            'chunk_size': chunk_size,
                            'dir_name': dir_name,
                            'example_batch_size': 1,
                            'make_process_example': make_example,
                            'kwargs': kwargs,
                        }
                        for i in range(num_chunks)
                    ]
                    results = pool.map(create_examples_from_file, process_args)

                # merge all results
                examples = [item for sublist in results for item in sublist]

                for file in chunk_file_paths:
                    os.remove(file)
            else:
                process_args = {
                    'in_file': path,
                    'chunk_size': max_examples,
                    'dir_name': dir_name,
                    'example_batch_size': 1,
                    'make_process_example': make_example,
                    'kwargs': kwargs,
                }
                examples = create_examples_from_file(process_args)

            if cache_input_data:
                os.makedirs(os.path.dirname(cache_name), exist_ok=True)
                logger.info(f'Caching data to {cache_name}')
                torch.save(examples, cache_name)

        super().__init__(examples, **kwargs)

    @classmethod
    def return_splits(cls, path, train='train', validation='eval', test='test', **kwargs):

        """Create dataset objects for splits of the ThingTalk dataset.
        Arguments:
            path: path to directory where data splits reside
            train: The prefix of the train data. Default: 'train'.
            validation: The prefix of the validation data. Default: 'eval'.
            test: The prefix of the test data. Default: 'test'.
            Remaining keyword arguments: Passed to the splits method of
                Dataset.
        """
        train_data = None if train is None else cls(os.path.join(path, train + '.tsv'), **kwargs)
        validation_data = None if validation is None else cls(os.path.join(path, validation + '.tsv'), **kwargs)
        test_data = None if test is None else cls(os.path.join(path, test + '.tsv'), **kwargs)

        aux_data = None
        do_curriculum = kwargs.get('curriculum', False)
        if do_curriculum:
            kwargs.pop('curriculum')
            aux_data = cls(os.path.join(path, 'aux' + '.tsv'), **kwargs)

        data_splits = Split(
            train=None if train is None else train_data,
            eval=None if validation is None else validation_data,
            test=None if test is None else test_data,
            aux=None if do_curriculum is False else aux_data,
        )

        all_paths = Split(
            train=None if train is None else os.path.join(path, train + '.tsv'),
            eval=None if validation is None else os.path.join(path, validation + '.tsv'),
            test=None if test is None else os.path.join(path, test + '.tsv'),
            aux=None if do_curriculum is False else os.path.join(path, 'aux' + '.tsv'),
        )

        return data_splits, all_paths



#
# Copyright (c) 2018, Salesforce, Inc.
#                     The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import copy
import json
import logging
import os
import shutil
from collections import defaultdict
from pprint import pformat

# multiprocessing with CUDA
from torch.multiprocessing import Process, set_start_method

from .data_utils.bootleg import Bootleg, extract_features_with_annotator, init_bootleg_annotator
from .run_bootleg import bootleg_process_splits

try:
    set_start_method('spawn')
except RuntimeError:
    pass

import torch

from . import models
from .arguments import check_and_update_generation_args
from .calibrate import ConfidenceEstimator
from .tasks.registry import get_tasks
from .util import (
    combine_folders_on_disk,
    get_devices,
    get_part_path,
    load_config_json,
    log_model_size,
    make_data_loader,
    set_seed,
    split_folder_on_disk,
)
from .validate import calculate_and_reduce_metrics, generate_with_model

logger = logging.getLogger(__name__)


def prepare_data(args, device):
    # initialize bootleg
    bootleg = None
    if args.do_ned and args.ned_retrieve_method == 'bootleg':
        bootleg = Bootleg(args)

    datasets = []
    paths = []
    if len(args.pred_src_languages) == 1 and len(args.tasks) > 1:
        args.pred_src_languages *= len(args.tasks)
    for i, task in enumerate(args.tasks):
        task_languages = args.pred_src_languages[i]
        logger.info(f'Loading {task}')
        kwargs = {'train': None, 'validation': None, 'test': None}
        if args.evaluate == 'train':
            del kwargs['train']  # deleting keys means use the default file name
        elif args.evaluate == 'valid':
            kwargs['validation'] = args.pred_set_name
        elif args.evaluate == 'test':
            del kwargs['test']
        else:
            raise ValueError('Split used for prediction should be either train, valid or test')

        kwargs.update(
            {
                'skip_cache': args.skip_cache,
                'subsample': args.subsample,
                'cached_path': os.path.join(args.cache, task.name),
                'all_dirs': task_languages,
                'almond_lang_as_question': args.almond_lang_as_question,
                'num_workers': args.num_workers,
                'separate_eval': args.separate_eval,
                'translate_no_answer': args.translate_no_answer,
                'ner_domains': args.ner_domains,
                'hf_test_overfit': args.hf_test_overfit,
            }
        )

        task_splits, task_paths = task.get_splits(root=args.data, lower=args.lower, **kwargs)
        if not isinstance(task_splits, list):
            task_splits = [task_splits]
            task_paths = [task_paths]
        task_data_processed = []
        task_path_processed = []
        for split, path in zip(task_splits, task_paths):
            assert (split.eval or split.test or split.train) and not split.aux
            if split.train:
                data = split.train
                path = path.train
            elif split.eval:
                data = split.eval
                path = path.eval
            else:
                data = split.test
                path = path.test
            if bootleg:
                if split.train or split.eval:
                    bootleg_process_splits(args, data.examples, path, task, bootleg)
                else:
                    # no prepped bootleg features are available
                    # extract features on-the-fly using bootleg annotator
                    bootleg_annotator = init_bootleg_annotator(args, device)
                    extract_features_with_annotator(data.examples, bootleg_annotator, args, task)
            task_data_processed.append(data)
            task_path_processed.append(path)
        datasets.append(task_data_processed)
        paths.append(task_path_processed)

    return datasets


def prepare_data_iterators(args, val_sets, numericalizer, device):
    logger.info('Preparing data iterators')
    if len(args.val_batch_size) == 1 and len(val_sets) > 1:
        args.val_batch_size *= len(val_sets)
    iters = []
    task_index = 0
    for task, bs, val_set in zip(args.tasks, args.val_batch_size, val_sets):
        task_iter = []
        task_languages = args.pred_src_languages[task_index]
        if task_languages is not None and args.separate_eval:
            task_languages = task_languages.split('+')
            assert len(task_languages) == len(val_set)
            for index, set_ in enumerate(val_set):
                loader, original_order = make_data_loader(
                    set_, numericalizer, bs, device, train=False, return_original_order=True
                )
                task_iter.append((task, task_languages[index], loader, original_order))
        # single language task or no separate eval
        else:
            loader, original_order = make_data_loader(
                val_set[0], numericalizer, bs, device, train=False, return_original_order=True
            )
            task_iter.append((task, task_languages, loader, original_order))

        iters.extend(task_iter)
        task_index += 1

    return iters


def run(args, device):

    # TODO handle multiple languages
    src_lang = args.pred_src_languages[0]
    tgt_lang = args.pred_tgt_languages[0]

    Model = getattr(models, args.model)
    model, _ = Model.load(
        args.path,
        model_checkpoint_file=args.checkpoint_name,
        args=args,
        device=device,
        tasks=args.tasks,
        src_lang=src_lang,
        tgt_lang=tgt_lang,
    )

    val_sets = prepare_data(args, device)
    model.add_new_vocab_from_data(args.tasks)

    iters = prepare_data_iterators(args, val_sets, model.numericalizer, device)

    log_model_size(logger, model, args.model)
    model.to(device)

    decaScore = []
    task_scores = defaultdict(list)
    model.eval()

    eval_dir = os.path.join(args.eval_dir, args.evaluate)
    os.makedirs(eval_dir, exist_ok=True)

    with torch.no_grad():
        for task, language, it, original_order in iters:
            logger.info(task.name)
            # single language task
            if language is None or 'multilingual' not in task.name:
                prediction_file_name = os.path.join(eval_dir, task.name + '.tsv')
                results_file_name = os.path.join(eval_dir, task.name + '.results.json')
            # multi language task
            else:
                prediction_file_name = os.path.join(eval_dir, task.name + '_{}.tsv'.format(language))
                results_file_name = os.path.join(eval_dir, task.name + '_{}.results.json'.format(language))
            if os.path.exists(prediction_file_name):
                if args.overwrite:
                    logger.warning(f'{prediction_file_name} already exists -- overwriting **')
                else:
                    raise OSError(f'{prediction_file_name} already exists')
            if os.path.exists(results_file_name):
                if args.overwrite:
                    logger.warning(f'{results_file_name} already exists -- overwriting **')
                else:
                    raise OSError(f'{results_file_name} already exists')

            if args.calibrator_paths is not None:
                confidence_estimators = []
                for path in args.calibrator_paths:
                    estimator = ConfidenceEstimator.load(path)
                    confidence_estimators.append(estimator)
                    logger.info('Loading confidence estimator "%s" from %s', estimator.name, path)
            else:
                confidence_estimators = None
            with torch.cuda.amp.autocast(enabled=args.mixed_precision):
                generation_output = generate_with_model(
                    model,
                    it,
                    model.numericalizer,
                    task,
                    args,
                    original_order=original_order,
                    output_confidence_features=args.save_confidence_features,
                    confidence_estimators=confidence_estimators,
                    disable_progbar=False,
                )

            if args.save_confidence_features:
                torch.save(generation_output.confidence_features, args.confidence_feature_path)

            # write into file
            # TODO change to jsonl format
            with open(prediction_file_name, 'w' + ('' if args.overwrite else 'x')) as prediction_file:
                for i in range(len(generation_output.example_ids)):
                    line = (
                        generation_output.example_ids[i] + '\t' + '\t'.join(generation_output.predictions[i])
                    )  # all outputs separated by '\t'
                    if args.calibrator_paths is not None:
                        for score in generation_output.confidence_scores:
                            line += '\t' + str(score[i])
                    prediction_file.write(line + '\n')

            if len(generation_output.answers) > 0:
                metrics_to_compute = task.metrics
                if args.main_metric_only:
                    metrics_to_compute = [metrics_to_compute[0]]
                metrics = calculate_and_reduce_metrics(
                    generation_output.predictions, generation_output.answers, metrics_to_compute, args.reduce_metrics
                )

                with open(results_file_name, 'w' + ('' if args.overwrite else '+')) as results_file:
                    results_file.write(json.dumps(metrics) + '\n')

                if not args.silent:
                    for i, (c, p, a) in enumerate(
                        zip(generation_output.contexts, generation_output.predictions, generation_output.answers)
                    ):
                        log_string = f'\nContext {i+1}: {c}\nPrediction {i + 1} ({len(p)} outputs): {p}\nAnswer {i + 1}: {a}\n'
                        if args.calibrator_paths is not None:
                            log_string += f'Confidence {i+1} : '
                            for score in generation_output.confidence_scores:
                                log_string += f'{score[i]:.3f}, '
                            log_string += '\n'
                        logger.info(log_string)
                    logger.info(metrics)

                task_scores[task].append((len(generation_output.answers), metrics[task.metrics[0]]))

    for task in task_scores.keys():
        decaScore.append(
            sum([length * score for length, score in task_scores[task]]) / sum([length for length, score in task_scores[task]])
        )

    logger.info('Evaluated Tasks:\n')
    for i, task in enumerate(args.tasks):
        logger.info(f'{task.name}: {decaScore[i]}')
    logger.info('-------------------')
    logger.info(f'DecaScore:  {sum(decaScore)}\n')
    logger.info(f'\nSummary: | {sum(decaScore)} | {" | ".join([str(x) for x in decaScore])} |\n')


def parse_argv(parser):
    parser.add_argument('--path', type=str, required=True, help='Folder to load the model from')
    parser.add_argument(
        '--evaluate',
        type=str,
        required=True,
        choices=['train', 'valid', 'test'],
        help='Which dataset to do predictions for (train, dev or test)',
    )
    parser.add_argument(
        '--pred_set_name',
        default='eval',
        type=str,
        help='Name of dataset to run prediction for; will be ignored if --evaluate is test',
    )
    parser.add_argument('--tasks', dest='task_names', nargs='+', help='task names for prediction')
    parser.add_argument(
        '--devices',
        default=None,
        nargs='+',
        type=int,
        help='a list of devices that can be used for prediction. By default, all devices will be used.',
    )
    parser.add_argument('--seed', default=123, type=int, help='Random seed.')
    parser.add_argument('--data', default='.data/', type=str, help='where to load data from.')
    parser.add_argument('--embeddings', default='.embeddings/', type=str, help='where to save embeddings.')
    parser.add_argument(
        '--checkpoint_name', default='best.pth', help='Checkpoint file to use (relative to --path, defaults to best.pth)'
    )
    parser.add_argument('--overwrite', action='store_true', help='whether to overwrite previously written predictions')
    parser.add_argument('--silent', action='store_true', help='whether to print predictions to stdout')

    parser.add_argument('--skip_cache', action='store_true', help='whether use exisiting cached splits or generate new ones')
    parser.add_argument('--eval_dir', type=str, required=True, help='use this directory to store eval results')
    parser.add_argument('--cache', default='.cache', type=str, help='where to save cached files')
    parser.add_argument('--subsample', default=20000000, type=int, help='subsample the eval/test datasets')

    parser.add_argument(
        '--pred_languages',
        type=str,
        nargs='+',
        dest='pred_src_languages',
        default=['en'],
        help='Specify dataset source languages used during prediction for multilingual tasks'
        'multiple languages for each task should be concatenated with +',
    )
    parser.add_argument(
        '--pred_tgt_languages',
        type=str,
        nargs='+',
        default=['en'],
        help='Specify dataset target languages used during prediction for multilingual tasks'
        'multiple languages for each task should be concatenated with +',
    )

    parser.add_argument('--separate_eval', action='store_true', help='evaluate on each language eval set separately')

    parser.add_argument(
        '--main_metric_only', action='store_true', help='If True, we only calculate the deca score metric for each task.'
    )
    # If not None, these values will override the values saved in the trained model's config file
    parser.add_argument(
        '--val_batch_size',
        nargs='+',
        default=None,
        type=int,
        help='Batch size for validation corresponding to tasks in val tasks',
    )
    parser.add_argument(
        "--reduce_metrics",
        type=str,
        default='max',
        choices=['max'],
        help='How to calculate the metric when there are multiple outputs per input.',
    )

    # These are generation hyperparameters. Each one can be a list of values in which case, we generate `num_outputs` outputs for each set of hyperparameters.
    parser.add_argument("--num_outputs", type=int, nargs='+', default=[1], help='number of sequences to output per input')
    parser.add_argument("--temperature", type=float, nargs='+', default=[0.0], help="temperature of 0 implies greedy sampling")
    parser.add_argument(
        "--repetition_penalty",
        type=float,
        nargs='+',
        default=[1.0],
        help="primarily useful for CTRL model; in that case, use 1.2",
    )
    parser.add_argument("--top_k", type=int, nargs='+', default=[0], help='0 disables top-k filtering')
    parser.add_argument("--top_p", type=float, nargs='+', default=[1.0], help='1.0 disables top-p filtering')
    parser.add_argument("--num_beams", type=int, nargs='+', default=[1], help='1 disables beam seach')
    parser.add_argument("--num_beam_groups", type=int, nargs='+', default=[1], help='1 disables diverse beam seach')
    parser.add_argument("--diversity_penalty", type=float, nargs='+', default=[0.0], help='0 disables diverse beam seach')
    parser.add_argument(
        "--no_repeat_ngram_size",
        type=int,
        nargs='+',
        default=[0],
        help='ngrams of this size cannot be repeated in the output. 0 disables it.',
    )
    parser.add_argument('--max_output_length', default=150, type=int, help='maximum output length for generation')

    # These are used for confidence calibration
    parser.add_argument(
        '--calibrator_paths',
        type=str,
        nargs='+',
        default=None,
        help='Can be a list. If provided, each calibrator will be used to output confidence scores for each prediction.',
    )
    parser.add_argument(
        '--save_confidence_features',
        action='store_true',
        help='If provided, will be used to output confidence scores for each prediction.',
    )
    parser.add_argument(
        "--confidence_feature_path", type=str, default=None, help='A .pkl file to save confidence features in.'
    )
    parser.add_argument(
        "--mc_dropout_num",
        type=int,
        default=0,
        help='Number of samples to use for Monte Carlo (MC) dropout. 0 disables MC dropout.',
    )
    parser.add_argument(
        "--override_confidence_labels",
        type=str,
        default=None,
        help='If provided, examples with this gold answer are marked as 1, and others as 0. Useful for out-of-domain detection.',
    )

    parser.add_argument(
        '--database_dir', type=str, help='Path to folder containing all files (e.g. alias2qids, pretrained models for bootleg)'
    )

    parser.add_argument(
        "--mixed_precision",
        action='store_true',
        help='If True, will use mixed precision for prediction.'
        'This reduces memory consumption and is especially faster on GPUs like NVIDIA V100 and T4. May slightly change the generated output.',
    )

    # TODO Update other tasks to use this argument too; so we can use predict for pure text generation (i.e. without reporting accuracy metrics)
    parser.add_argument(
        '--translate_no_answer',
        action='store_true',
        help='if true the provided dataset would not contain the answer (translated sentence)',
    )
    parser.add_argument('--plot_heatmaps', action='store_true', help='whether to plot cross-attention heatmaps')


def set_default_values(args):
    """
    sets default values that depend on other input arguments
    """
    if args.confidence_feature_path is None:
        args.confidence_feature_path = os.path.join(args.path, 'confidence_features.pkl')


def check_args(args):

    if len(args.task_names) != len(args.pred_src_languages):
        raise ValueError(
            'You have to define prediction languages for each task'
            'Use None for single language tasks. Also provide languages in the same order you provided the tasks.'
        )

    if getattr(args, 'do_ned', False) and getattr(args, 'ned_retrieve_method', None) == 'bootleg':
        with open(os.path.join(args.path, 'config.json')) as config_file:
            config = json.load(config_file)
        if args.subsample > config['subsample']:
            raise ValueError('To use bootleg, you have to use a subsample value less than the number of prepped examples.')


def main(args):
    load_config_json(args)
    check_and_update_generation_args(args)
    check_args(args)
    set_default_values(args)

    set_seed(args)
    args.tasks = list(get_tasks(args.task_names, args).values())

    logger.info(f'Arguments:\n{pformat(vars(args))}')
    logger.info(f'Loading from {args.best_checkpoint}')

    devices = get_devices(args.devices)

    if len(devices) > 1:
        logger.info(f'Independent multi-GPU generation on following devices: {devices}')
        all_processes = []
        all_data_folders = split_folder_on_disk(args.data, len(devices))

        for device_id in range(len(devices)):
            copy_args = copy.copy(args)
            copy_args.data = all_data_folders[device_id]
            copy_args.eval_dir = get_part_path(args.eval_dir, device_id)

            p = Process(target=run, args=(copy_args, devices[device_id]))
            all_processes.append(p)
            p.start()

        for p in all_processes:
            p.join()

        for folder in all_data_folders:
            shutil.rmtree(folder)
        combine_folders_on_disk(args.eval_dir, len(devices), line_group_size=1, delete=True)

    else:
        logger.info(f'Single device generation on: {devices[0]}')
        run(args, devices[0])



#
# Copyright (c) 2021 The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging

from transformers import AutoConfig, AutoModelForTokenClassification

from ..data_utils.numericalizer import TransformerNumericalizer
from ..models.base import GenieModel
from ..util import adjust_language_code

logger = logging.getLogger(__name__)


class TransformerForTokenClassification(GenieModel):
    def __init__(self, config=None, *inputs, args, tasks, vocab_sets, save_directory=None, **kwargs):

        num_labels = 0
        if args.num_labels is not None:
            num_labels = args.num_labels
        else:
            for task in tasks:
                # if having multiple tasks choose max num_labels
                if hasattr(task, 'num_labels'):
                    num_labels = max(num_labels, task.num_labels)

        config = AutoConfig.from_pretrained(
            args.pretrained_model, cache_dir=args.embeddings, num_labels=num_labels, finetuning_task='ned'
        )
        GenieModel.__init__(self, config)
        self.args = args
        if hasattr(config, 'd_model'):
            args.dimension = config.d_model
        else:
            args.dimension = config.hidden_size

        self.src_lang, self.tgt_lang = adjust_language_code(
            config, args.pretrained_model, kwargs.get('src_lang', 'en'), kwargs.get('tgt_lang', 'en')
        )

        if save_directory is not None:
            self.model = AutoModelForTokenClassification.from_config(config)
        else:
            self.model = AutoModelForTokenClassification.from_pretrained(
                self.args.pretrained_model, cache_dir=self.args.embeddings, config=config
            )

        self.numericalizer = TransformerNumericalizer(
            self.args.pretrained_model,
            args,
            max_generative_vocab=None,
            save_dir=save_directory,
            config=config,
            src_lang=self.src_lang,
            tgt_lang=self.tgt_lang,
            vocab_sets=vocab_sets,
            tasks=tasks,
        )

        self.model.resize_token_embeddings(self.numericalizer.num_tokens)

        self.numericalizer.answer_pad_id = -100

    def add_new_vocab_from_data(self, tasks, resize_decoder=False):
        super().add_new_vocab_from_data(tasks, resize_decoder)
        self.model.resize_token_embeddings(self.numericalizer.num_tokens)

    def forward(self, *input, **kwargs):
        if self.training:
            batch = input[0]
            outputs = self.model(
                batch.context.value,
                labels=batch.answer.value,
                attention_mask=(batch.context.value != self.numericalizer.pad_id),
            )
            return outputs
        else:
            return self.model(**kwargs)



#
# Copyright (c) 2018, The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''
Created on Mar 3, 2019

@author: gcampagn
'''

import json
import logging
import os

import torch

logger = logging.getLogger(__name__)


class Saver(object):
    '''
    Wrap pytorch's save functionality into an interface similar to tensorflow.train.Saver

    In particular, this class takes care of automatically cleaning up old checkpoints,
    and creating checkpoint files to keep track of which saves are valid and which are not.
    '''

    def __init__(self, savedir, max_to_keep=5):
        self._savedir = savedir
        self._max_to_keep = max_to_keep
        assert max_to_keep >= 1

        self._loaded_last_checkpoints = False
        self._latest_checkpoint = None
        self._all_checkpoints = None

    def _maybe_load_last_checkpoints(self):
        if self._loaded_last_checkpoints:
            return

        try:
            with open(os.path.join(self._savedir, 'checkpoint.json')) as fp:
                data = json.load(fp)
                self._loaded_last_checkpoints = True
                self._all_checkpoints = data['all']
                self._latest_checkpoint = data['latest']
        except FileNotFoundError:
            self._loaded_last_checkpoints = True
            self._all_checkpoints = []
            self._latest_checkpoint = None

    def save(self, save_model_state_dict, save_opt_state_dict, global_step):
        self._maybe_load_last_checkpoints()

        model_name = 'iteration_' + str(global_step) + '.pth'
        opt_name = 'iteration_' + str(global_step) + '_optim.pth'

        self._latest_checkpoint = model_name
        self._all_checkpoints.append(model_name)
        if len(self._all_checkpoints) > self._max_to_keep:
            try:
                todelete = self._all_checkpoints.pop(0)
                os.unlink(os.path.join(self._savedir, todelete))
                opt_todelete = todelete.rsplit('.', 1)[0] + '_optim.' + todelete.rsplit('.', 1)[1]
                os.unlink(os.path.join(self._savedir, opt_todelete))
            except (OSError, IOError) as e:
                logging.warning('Failed to delete old checkpoint: %s', e)
        torch.save(save_model_state_dict, os.path.join(self._savedir, model_name))
        torch.save(save_opt_state_dict, os.path.join(self._savedir, opt_name))
        with open(os.path.join(self._savedir, 'checkpoint.json'), 'w') as fp:
            json.dump(dict(all=self._all_checkpoints, latest=self._latest_checkpoint), fp)



#
# Copyright (c) 2019, The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import logging

from .base_task import BaseTask

logger = logging.getLogger(__name__)


class TaskRegistry:
    """
    A container for all the tasks supported by the library
    """

    def __init__(self):
        self.store = dict()

    def __setitem__(self, name, cls):
        if name in self.store:
            raise ValueError(f'Duplicate task {name}')
        self.store[name] = cls

    def __getitem__(self, name):
        if name not in self.store:
            logger.warning(f'Unrecognized task {name}, using generic code')
            self.store[name] = BaseTask
        return self.store[name]

    def __len__(self):
        return len(self.store)

    def __iter__(self):
        return iter(self.store)


_registry = TaskRegistry()
_registry['generic'] = BaseTask


def task_name_to_cls_name(name):
    return name.split('.')[0]


def register_task(name):
    def decorator(cls):
        _registry[name] = cls
        return cls

    return decorator


def get_tasks(names, args, available_tasks=None):
    tasks = dict()
    for name in names:
        if available_tasks and name in available_tasks:
            tasks[name] = available_tasks[name]
        else:
            tasks[name] = _registry[task_name_to_cls_name(name)](name, args)
    return tasks



# Copyright 2021 The Board of Trustees of the Leland Stanford Junior University
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#  contributors may be used to endorse or promote products derived from
#  this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import json
import logging

logger = logging.getLogger(__name__)


def parse_argv(parser):
    parser.add_argument('--eval_file', type=str, required=True, help='path to eval json file')


def main(args):
    with open(args.eval_file, 'r') as fin:
        metrics = json.load(fin)

    # kf dashboard is hardcoded to only show two metrics
    # we choose the first two from task metrics that are important
    items = list(metrics.items())[:2]

    extracted_metrics = [{'name': key, 'numberValue': val / 100, 'format': "PERCENTAGE"} for key, val in items]

    metrics = {'metrics': extracted_metrics}
    with open('/tmp/mlpipeline-metrics.json', 'w') as fout:
        json.dump(metrics, fout)



#
# Copyright (c) 2018, Salesforce, Inc.
#                     The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import logging
import logging.handlers
import math
import os
import time
from copy import deepcopy
from functools import partial
from pprint import pformat

import numpy as np
import torch
from tensorboardX import SummaryWriter
from transformers import (
    AdamW,
    get_constant_schedule_with_warmup,
    get_cosine_schedule_with_warmup,
    get_linear_schedule_with_warmup,
)

from . import arguments, models
from .arguments import save_args
from .data_utils.bootleg import Bootleg
from .model_utils.parallel_utils import NamedTupleCompatibleDataParallel
from .model_utils.saver import Saver
from .run_bootleg import bootleg_process_splits
from .util import (
    elapsed_time,
    get_devices,
    get_trainable_params,
    log_model_size,
    make_data_loader,
    ned_dump_entity_type_pairs,
    set_seed,
)
from .validate import print_results, validate


def initialize_logger(args):
    # set up file logger
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    handler = logging.handlers.RotatingFileHandler(
        os.path.join(args.log_dir, 'train.log'), maxBytes=1024 * 1024 * 10, backupCount=1
    )
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(name)s - %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    handler.setLevel(logging.DEBUG)
    logger.addHandler(handler)
    logger.propagate = False

    return logger


def prepare_data(args, logger):

    # initialize bootleg
    bootleg = None
    if args.do_ned and args.ned_retrieve_method == 'bootleg':
        bootleg = Bootleg(args)

    train_sets, val_sets, aux_sets = [], [], []

    train_eval_shared_kwargs = {
        'subsample': args.subsample,
        'skip_cache': args.skip_cache,
        'cache_input_data': args.cache_input_data,
        'sentence_batching': args.sentence_batching,
        'almond_lang_as_question': args.almond_lang_as_question,
        'num_workers': args.num_workers,
    }

    if any(args.train_iterations):
        for task in args.train_tasks:
            logger.info(f'Loading {task.name}')
            kwargs = {'test': None, 'validation': None}
            kwargs.update(train_eval_shared_kwargs)
            kwargs['all_dirs'] = args.train_src_languages
            kwargs['cached_path'] = os.path.join(args.cache, task.name)
            kwargs['ner_domains'] = args.ner_domains
            if args.use_curriculum:
                kwargs['curriculum'] = True

            logger.info(f'Adding {task.name} to training datasets')
            t0 = time.time()
            splits, paths = task.get_splits(args.data, lower=args.lower, **kwargs)

            t1 = time.time()
            logger.info('Data loading took {:.2f} seconds'.format(t1 - t0))
            assert not splits.eval and not splits.test
            if args.use_curriculum:
                assert splits.aux
                aux_sets.append(splits.aux)
                logger.info(f'{task.name} has {len(splits.aux)} auxiliary examples')
            else:
                assert splits.train

            if bootleg:
                bootleg_process_splits(args, splits.train.examples, paths.train, task, bootleg)

            train_sets.append(splits.train)
            logger.info(f'{task.name} has {len(splits.train)} training examples')

            logger.info(f"train all_schema_types: {getattr(task, 'all_schema_types', None)}")

            if task.name.startswith('almond'):
                if args.ned_features_default_val:
                    args.db_unk_id = int(args.ned_features_default_val[0])
                else:
                    args.db_unk_id = 0
                if args.do_ned:
                    if bootleg:
                        args.num_db_types = len(bootleg.typeqid2id)
                    elif getattr(task, 'db', None):
                        args.num_db_types = len(task.db.typeqid2id)
                else:
                    args.num_db_types = 0
            else:
                args.db_unk_id = 0
                args.num_db_types = 0
            save_args(args, force_overwrite=True)

        for task in args.val_tasks:
            logger.info(f'Loading {task.name}')
            kwargs = {'train': None, 'test': None}
            # choose best model based on this dev set
            if args.eval_set_name is not None:
                kwargs['validation'] = args.eval_set_name
            kwargs.update(train_eval_shared_kwargs)
            kwargs['all_dirs'] = args.eval_src_languages
            kwargs['cached_path'] = os.path.join(args.cache, task.name)
            kwargs['ner_domains'] = args.ner_domains
            kwargs['hf_test_overfit'] = args.hf_test_overfit

            logger.info(f'Adding {task.name} to validation datasets')
            splits, paths = task.get_splits(args.data, lower=args.lower, **kwargs)

            assert not splits.train and not splits.test and not splits.aux
            logger.info(f'{task.name} has {len(splits.eval)} validation examples')

            if bootleg:
                bootleg_process_splits(args, splits.eval.examples, paths.eval, task, bootleg)

            val_sets.append(splits.eval)

            logger.info(f"validation all_schema_types: {getattr(task, 'all_schema_types', None)}")

    return train_sets, val_sets, aux_sets


accumulated_batch_lengths = 0


def train_step(model, batch, iteration, opt, devices, lr_scheduler=None, grad_clip=None, gradient_accumulation_steps=1):
    # Since the batch size is different in each call to this function due to dynamic batching, we need to keep track of
    # the total batch size
    global accumulated_batch_lengths
    model.train()
    if (iteration) % gradient_accumulation_steps == 0:
        opt.zero_grad()
    loss = model(batch).loss
    if torch.isnan(loss).any():
        raise RuntimeError('Got NaN loss %s', str(loss))
    if len(devices) > 1:
        loss = loss.mean()
    non_accumulated_loss = loss.item()
    loss = loss * len(batch[0])
    accumulated_batch_lengths += len(batch[0])

    loss.backward()
    grad_norm = None
    if (iteration + 1) % gradient_accumulation_steps == 0:
        for p in model.parameters():
            if p.grad is None:
                continue
            p.grad /= accumulated_batch_lengths
        accumulated_batch_lengths = 0
        if grad_clip > 0.0:
            grad_norm = torch.nn.utils.clip_grad_norm_(model.params, grad_clip)
        opt.step()
        lr_scheduler.step()

    return non_accumulated_loss, grad_norm


def update_fraction(args, task_iteration):
    if args.curriculum_strategy == 'linear':
        next_fraction = args.curriculum_rate * task_iteration
    elif args.curriculum_strategy == 'exp':
        next_fraction = args.curriculum_rate * np.exp(task_iteration)

    fraction = min(args.curriculum_max_frac, next_fraction)

    return fraction


def should_validate(iteration, val_every, resume, start_iteration):
    if val_every is None:
        return False
    return (iteration % val_every == 0) or (resume and iteration == start_iteration)


def should_save(iteration, save_every):
    if save_every is None:
        return False
    return iteration % save_every == 0


def should_log(iteration, log_every):
    if log_every is None:
        return False
    return iteration % log_every == 0


def do_validate(
    iteration, args, model, numericalizer, val_iters, *, train_task, round_progress, task_progress, writer, logger
):
    deca_score = 0
    for val_task_idx, (val_task, val_iter) in enumerate(val_iters):
        output, metric_dict = validate(val_task, val_iter, model, numericalizer, args, num_print=args.num_print)
        val_loss = output.loss
        if val_loss is not None:
            log_entry = f'{args.timestamp}:{elapsed_time(logger)}:iteration_{iteration}:{round_progress}train_{train_task.name}:{task_progress}val_{val_task.name}:val_loss{val_loss.item():.4f}:'
            writer.add_scalar(f'loss/{val_task.name}/val', val_loss.item(), iteration)
        else:
            log_entry = f'{args.timestamp}:{elapsed_time(logger)}:iteration_{iteration}:{round_progress}train_{train_task.name}:{task_progress}val_{val_task.name}:'

        metric_entry = ''
        for metric_key, metric_value in metric_dict.items():
            metric_entry += f'{metric_key}_{metric_value:.2f}:'
        metric_entry = metric_entry[:-1]

        deca_score += metric_dict[val_task.metrics[0]]

        # val log
        logger.info(log_entry + metric_entry)
        if writer is not None:
            for metric_key, metric_value in metric_dict.items():
                writer.add_scalar(f'{val_task.name}/{metric_key}/val', metric_value, iteration)
    if writer is not None:
        writer.add_scalar('deca/val', deca_score, iteration)
    logger.info(
        f'{args.timestamp}:{elapsed_time(logger)}:iteration_{iteration}:{round_progress}train_{train_task.name}:{task_progress}val_deca:deca_{deca_score:.2f}'
    )

    return deca_score


def maybe_save(
    iteration,
    model,
    opt,
    deca_score,
    best_decascore,
    *,
    saver,
    logger,
    train_task,
    round_progress,
    task_progress,
    timestamp,
    log_dir,
    model_parallel,
):
    should_save_best = False
    if deca_score is not None and (best_decascore is None or best_decascore < deca_score):
        best_decascore = deca_score
        should_save_best = True

    # DataParallel and ModelParallel are mutually exclusive
    if model_parallel:
        model_state_dict = model.state_dict()
    else:
        # punch through the nn.DataParallel to access the real model, otherwise we won't be able
        # to load this model later
        model_state_dict = model.module.state_dict()

    model_state_dict = {k: v.cpu() for k, v in model_state_dict.items()}

    save_model_state_dict = {'model_state_dict': model_state_dict, 'best_decascore': best_decascore}
    save_opt_state_dict = opt.state_dict()
    save_opt_state_dict.update({'start_iteration': iteration})

    saver.save(save_model_state_dict, save_opt_state_dict, global_step=iteration)
    if should_save_best:
        logger.info(
            f'{timestamp}:{elapsed_time(logger)}:iteration_{iteration}:{round_progress}train_{train_task.name}:{task_progress}saving new best model'
        )
        torch.save(save_model_state_dict, os.path.join(log_dir, 'best.pth'))
        torch.save(save_opt_state_dict, os.path.join(log_dir, 'best_optim.pth'))

        if model_parallel:
            model.numericalizer.save(saver._savedir)
        else:
            model.module.numericalizer.save(saver._savedir)

    return best_decascore


def do_log_training_loss(
    iteration,
    loss,
    *,
    lr_scheduler,
    grad_norm,
    num_examples,
    len_contexts,
    len_answers,
    logger,
    train_task,
    round_progress,
    epochs,
    task_progress,
    timestamp,
    writer,
    log_prefix,
):
    avg_batch_size = f'avbatch_{num_examples:.0f}_{len_contexts:.0f}_{len_answers:.0f}:'
    logger.info(
        f'{timestamp}:{elapsed_time(logger)}:iteration_{iteration}:epoch_{epochs:.2f}:{round_progress}train_{train_task.name}:{task_progress}{avg_batch_size}{log_prefix}/loss_{loss:.4f}'
    )

    if writer is not None:
        writer.add_scalar(f'{log_prefix}/loss/{train_task.name}', loss, iteration)

        if lr_scheduler is not None:
            writer.add_scalar(f'{log_prefix}/lr', lr_scheduler.get_last_lr(), iteration)
        if grad_norm is not None:
            writer.add_scalar(f'{log_prefix}/norm', grad_norm, iteration)


def np_coin(prob):
    return np.random.uniform() < prob


def get_next_batch(train_iter, aux_iters, *, task, task_idx, task_fraction, use_curriculum):
    if use_curriculum:
        aux_iter = aux_iters[task_idx][1]
        prob = np_coin(task_fraction[task])
        if prob == 'aux':
            return next(aux_iter)
        else:
            assert prob == 'train'
            batch = next(train_iter)

    else:
        batch = next(train_iter)

    return batch


def train(
    args,
    devices,
    model,
    opt,
    lr_scheduler,
    train_sets,
    train_iterations,
    numericalizer,
    *,
    log_every,
    val_every,
    save_every,
    rounds,
    val_sets,
    aux_sets,
    writer,
    logger,
    log_prefix,
    start_iteration=1,
    rnd=1,
    best_decascore,
    use_curriculum,
):
    """main training function"""
    local_loss, num_examples, len_contexts, len_answers, iteration = 0, 0, 0, 0, 1

    train_iter_deep = deepcopy(train_iterations)

    task_iteration = dict()
    task_done = dict()
    task_fraction = dict()
    task_total_num_examples = dict()
    task_train_size = dict()  # number of examples in each task

    for task, train_set in zip(args.train_tasks, train_sets):
        task_iteration[task] = 1
        task_done[task] = False
        task_fraction[task] = 0.0
        task_total_num_examples[task] = 0.0
        task_train_size[task] = len(train_set)

    saver = Saver(args.log_dir, args.max_to_keep)
    per_task_iterations = 0

    logger.info('Preparing iterators')
    main_device = devices[0]

    t0 = time.time()
    train_iters = [
        (task, make_data_loader(dataset, numericalizer, tok, main_device, train=True))
        for task, dataset, tok in zip(args.train_tasks, train_sets, args.train_batch_tokens)
    ]
    t1 = time.time()
    logger.info('Preparing iterators took {:.2f} seconds'.format(t1 - t0))

    train_iters = [(task, iter(train_iter)) for task, train_iter in train_iters]
    # save memory
    del train_sets

    val_iters = [
        (task, make_data_loader(dataset, numericalizer, bs, main_device, train=False))
        for task, dataset, bs in zip(args.val_tasks, val_sets, args.val_batch_size)
    ]
    # save memory
    del val_sets

    aux_iters = []
    if use_curriculum:
        aux_iters = [
            (name, make_data_loader(dataset, numericalizer, tok, main_device, train=True))
            for name, dataset, tok in zip(args.train_tasks, aux_sets, args.train_batch_tokens)
        ]
        aux_iters = [(task, iter(aux_iter)) for task, aux_iter in aux_iters]
        # save memory
        del aux_sets

    zero_loss = 0
    logger.info(f'Begin {log_prefix}')

    if any(train_iterations):
        while not all(task_done.values()):
            # For some number of rounds, we 'jump start' some subset of the tasks
            # by training them and not others
            # once the specified number of rounds is completed,
            # switch to normal round robin training
            if rnd < args.jump_start:
                train_iterations = [0] * len(train_iterations)
                for j in range(args.n_jump_start):
                    train_iterations[j] = 1
            else:
                train_iterations = train_iter_deep

            for task_idx, (task, train_iter) in enumerate(train_iters):
                task_iterations = train_iterations[task_idx] if train_iterations is not None else None
                if task_iterations == 0:
                    continue
                if task_iterations is not None and task_iteration[task] > task_iterations:
                    task_done[task] = True
                    continue

                # load batches even if (args.resume == True) and we are going to skip the iteration
                # this makes runs that are resumed have the exact same behavior as runs that are
                # finished in one pass (given that the random seed is the same).
                batch = get_next_batch(
                    train_iter,
                    aux_iters,
                    task=task,
                    task_idx=task_idx,
                    task_fraction=task_fraction,
                    use_curriculum=use_curriculum,
                )

                if iteration < start_iteration:
                    # skip this iteration (this is done to ensure iterators are at the same position when resuming)
                    task_iteration[task] += 1
                    iteration += 1
                    if (iteration + 1) % args.gradient_accumulation_steps == 0:
                        lr_scheduler.step()  # update the learning rate
                    continue

                task_progress = f'{task_iteration[task]}/{task_iterations}:' if task_iterations is not None else ''
                round_progress = f'round_{rnd}:' if rounds else ''

                # param update
                loss, grad_norm = train_step(
                    model,
                    batch,
                    iteration,
                    opt,
                    devices,
                    lr_scheduler=lr_scheduler,
                    grad_clip=args.grad_clip,
                    gradient_accumulation_steps=args.gradient_accumulation_steps,
                )
                if loss is None:
                    logger.info('Encountered NAN loss during training. Continue training ignoring the current batch')
                    continue
                if loss < 1e-6:
                    zero_loss += 1
                    if zero_loss >= 100:
                        logger.info('Found loss less than 1e-6 for 100 steps, stopping.')
                        return
                else:
                    zero_loss = 0

                # update curriculum fraction
                if args.use_curriculum:
                    task_fraction[task] = update_fraction(args, task_iteration[task])

                # train metrics
                local_loss += loss

                # train logs
                num_examples += batch.context.value.size(0)
                len_contexts += batch.context.value.size(1)
                len_answers += batch.answer.value.size(1)

                task_total_num_examples[task] += batch.context.value.size(0)

                if should_log(iteration, log_every):
                    local_loss /= log_every
                    num_examples /= log_every
                    len_contexts /= log_every
                    len_answers /= log_every
                    do_log_training_loss(
                        iteration,
                        local_loss,
                        lr_scheduler=lr_scheduler,
                        grad_norm=grad_norm,
                        num_examples=num_examples,
                        len_contexts=len_contexts,
                        len_answers=len_answers,
                        logger=logger,
                        writer=writer,
                        train_task=task,
                        round_progress=round_progress,
                        epochs=task_total_num_examples[task] / task_train_size[task],
                        task_progress=task_progress,
                        timestamp=args.timestamp,
                        log_prefix=log_prefix,
                    )
                    num_examples = 0
                    len_contexts = 0
                    len_answers = 0
                    local_loss = 0

                # validate
                if should_validate(iteration, val_every, resume=args.resume, start_iteration=start_iteration):
                    if args.print_train_examples_too:
                        names = ['answer', 'context']
                        values = [
                            numericalizer.reverse(batch.answer.value.data, 'answer'),
                            numericalizer.reverse(batch.context.value.data, 'context'),
                        ]
                        num_print = min(len(values[0]), args.num_print)
                        print_results(names, values, num_print=num_print)

                    deca_score = do_validate(
                        iteration,
                        args,
                        model,
                        numericalizer,
                        val_iters,
                        train_task=task,
                        round_progress=round_progress,
                        task_progress=task_progress,
                        writer=writer,
                        logger=logger,
                    )

                    # saving
                    if should_save(iteration, save_every):
                        best_decascore = maybe_save(
                            iteration,
                            model,
                            opt,
                            deca_score,
                            best_decascore,
                            saver=saver,
                            logger=logger,
                            train_task=task,
                            round_progress=round_progress,
                            task_progress=task_progress,
                            timestamp=args.timestamp,
                            log_dir=args.log_dir,
                            model_parallel=args.model_parallel,
                        )

                # book keeping
                task_iteration[task] += 1
                iteration += 1

            # book keeping
            per_task_iterations += 1
            rnd += 1

        logger.info(f'{log_prefix} is done after {per_task_iterations - 1} iterations')

    else:
        # Save pretrained models as is without any finetuning
        # Useful for doing prediction/ generation on those models with genienlp
        for task in args.train_tasks:
            maybe_save(
                0,
                model,
                opt,
                deca_score=0,
                best_decascore=-1,
                saver=saver,
                logger=logger,
                train_task=task,
                round_progress=0,
                task_progress=0,
                timestamp=args.timestamp,
                log_dir=args.log_dir,
                model_parallel=args.model_parallel,
            )

        logger.info(f'{args.pretrained_model} model is saved to {args.save} without any fine-tuning')


def get_transformer_learning_rate(i, *, dimension, warmup):
    i += 1
    return 1.0 / math.sqrt(dimension) * min(1 / math.sqrt(i), i / (warmup * math.sqrt(warmup)))


def get_sgd_learning_rate(i, *, warmup):
    i += 1
    return min(math.sqrt(warmup) / math.sqrt(i), i / warmup)


def init_opt(args, model, logger):
    if args.optimizer == 'adam':
        # Adam with transformer schedule has a different set of default hyperparameters:
        if args.lr_schedule == 'transformer':
            opt = torch.optim.Adam(
                model.params, lr=args.lr_multiply, betas=(0.9, 0.98), eps=1e-9, weight_decay=args.weight_decay
            )
        else:
            opt = torch.optim.Adam(
                model.params, lr=args.lr_multiply, betas=(args.beta0, 0.999), weight_decay=args.weight_decay
            )
    elif args.optimizer == 'adamw':
        opt = AdamW(model.params, lr=args.lr_multiply, weight_decay=args.weight_decay)
    elif args.optimizer == 'radam':
        import radam

        if args.warmup > 1:
            logger.warning('With RAdam optimizer, warmup is never applied')
        opt = radam.RAdam(model.params, lr=args.lr_multiply, betas=(args.beta0, 0.999), weight_decay=args.weight_decay)
    else:
        assert args.optimizer == 'sgd'
        opt = torch.optim.SGD(model.params, lr=args.lr_multiply, weight_decay=args.weight_decay)

    if args.lr_schedule == 'transformer':
        lr_lambda = partial(get_transformer_learning_rate, dimension=args.dimension, warmup=args.warmup)
        scheduler = torch.optim.lr_scheduler.LambdaLR(opt, lr_lambda)
    elif args.lr_schedule == 'constant':
        scheduler = get_constant_schedule_with_warmup(opt, num_warmup_steps=args.warmup)
    elif args.lr_schedule == 'linear':
        scheduler = get_linear_schedule_with_warmup(
            opt,
            num_training_steps=sum(args.train_iterations) // args.gradient_accumulation_steps,
            num_warmup_steps=args.warmup,
        )
    elif args.lr_schedule == 'cosine':
        scheduler = get_cosine_schedule_with_warmup(
            opt,
            num_training_steps=sum(args.train_iterations) // args.gradient_accumulation_steps,
            num_warmup_steps=args.warmup,
            num_cycles=0.5,
        )
    elif args.lr_schedule == 'sgd':
        lr_lambda = partial(get_sgd_learning_rate, warmup=args.warmup)
        scheduler = torch.optim.lr_scheduler.LambdaLR(opt, lr_lambda)
    else:
        raise ValueError('Invalid learning rate scheduler.')

    return opt, scheduler


def main(args):
    args = arguments.post_parse_general(args)
    args = arguments.post_parse_train_specific(args)
    if args is None:
        return

    set_seed(args)
    devices = get_devices(args.devices)
    logger = initialize_logger(args)
    logger.info(f'Arguments:\n{pformat(vars(args))}')

    model_name = args.model
    model_class = getattr(models, model_name)

    tasks = set(args.train_tasks) | set(args.val_tasks)
    train_sets, val_sets, aux_sets = prepare_data(args, logger)

    if (args.use_curriculum and aux_sets is None) or (not args.use_curriculum and len(aux_sets) > 0):
        logging.error('Something unpleasant is happening with curriculum')

    logger.info('Processing')
    logger.start = time.time()

    # TODO handle multiple languages
    # TODO handle different train and eval languages
    src_lang = args.train_src_languages.split('+')[0]
    tgt_lang = args.train_tgt_languages.split('+')[0]

    ########## initialize model
    best_decascore = None
    if args.load is not None:
        model, best_decascore = model_class.load(
            args.save,
            args=args,
            model_checkpoint_file=args.load,
            vocab_sets=train_sets + val_sets,
            tasks=tasks,
            device=devices[0],
            src_lang=src_lang,
            tgt_lang=tgt_lang,
        )
        model.add_new_vocab_from_data(tasks=tasks, resize_decoder=True)
        if not args.resume:
            # we are fine-tuning, so reset the best score since the new fine-tune dataset usually has a different validation set from the original
            best_decascore = None
    else:
        logger.info(f'Initializing a new {model_name}')
        model = model_class(args=args, vocab_sets=train_sets + val_sets, tasks=tasks, src_lang=src_lang, tgt_lang=tgt_lang)

    # dump entities if required
    if args.ned_dump_entity_type_pairs and args.add_types_to_text == 'append':
        for task, train_set, val_set in zip(tasks, train_sets, val_sets):
            ned_dump_entity_type_pairs(train_set, args.data, 'train', task.utterance_field())
            ned_dump_entity_type_pairs(val_set, args.data, 'eval', task.utterance_field())

    params = get_trainable_params(model)
    log_model_size(logger, model, model_name)

    if args.model_parallel:
        n_layers = len(model.model.encoder.block)
        layers = list(range(n_layers))

        if args.mp_device_ratio is not None:
            if len(args.devices) > torch.cuda.device_count() or max(args.devices) >= torch.cuda.device_count():
                raise ValueError('Provided GPU devices exceeds the number of available GPU')

            mp_device_ratio = [(device_ratio / sum(args.mp_device_ratio)) for device_ratio in args.mp_device_ratio]
            layers_list = []
            beg = 0
            for i in range(len(args.devices)):
                end = min(beg + math.ceil(n_layers * mp_device_ratio[i]), n_layers)
                layers_list.append(layers[beg:end])
                beg = end

            if any([len(val_list) == 0 for val_list in layers_list]):
                raise ValueError(
                    'One device got no layers given the mp_device_ratio you provided'
                    'Hint: if you do not provide mp_device_ratio, layers will be distributed evenly'
                )

        else:
            n_blocks = int(math.ceil(n_layers / len(args.devices)))
            layers_list = list(layers[i : i + n_blocks] for i in range(0, n_layers, n_blocks))

        device_map = dict(zip(args.devices, layers_list))
        model.model.parallelize(device_map)
        logger.info(f'Model parallel is used with following device map: {model.model.device_map}')
    else:
        model.to(devices[0])
        model = NamedTupleCompatibleDataParallel(model, device_ids=devices)
    model.params = params
    ##########

    opt, lr_scheduler = init_opt(args, model, logger)
    start_iteration = 1

    if args.resume:
        logger.info(f'Resuming training from {os.path.splitext(args.load)[0]}_optim.pth')
        # load optimizer's state_dict to cpu first to avoid GPU memory surge. Will crash with OOM if `map_location='cpu'` is not specified.
        opt_state_dict = torch.load(os.path.join(args.save, f'{os.path.splitext(args.load)[0]}_optim.pth'), map_location='cpu')
        start_iteration = opt_state_dict.pop('start_iteration')
        logger.info(f'Starting iteration is {start_iteration}')
        opt.load_state_dict(opt_state_dict)

    if hasattr(args, 'tensorboard') and args.tensorboard:
        logger.info('Initializing Writer')
        writer = SummaryWriter(log_dir=args.tensorboard_dir, purge_step=start_iteration, flush_secs=60)
    else:
        writer = None

    train(
        args,
        devices,
        model,
        opt,
        lr_scheduler,
        train_sets,
        args.train_iterations,
        model.module.numericalizer if not args.model_parallel else model.numericalizer,
        val_sets=val_sets,
        aux_sets=aux_sets,
        logger=logger,
        writer=writer,
        log_every=args.log_every,
        val_every=args.val_every,
        save_every=args.save_every,
        rounds=len(train_sets) > 1,
        start_iteration=start_iteration,
        use_curriculum=args.use_curriculum,
        best_decascore=best_decascore,
        log_prefix='training',
    )

    if writer is not None:
        writer.close()  # otherwise the last written value may not be flushed



#
# Copyright (c) 2021, Salesforce, Inc.
#                     The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging

import kfserving

from .server import Server, init
from .util import log_model_size

logger = logging.getLogger(__name__)


class KFModelServer(kfserving.KFModel):
    def __init__(
        self, name, args, numericalizer, model, device, confidence_estimators, estimator_filenames, bootleg_annotator
    ):
        super().__init__(name)
        self.server = Server(args, numericalizer, model, device, confidence_estimators, estimator_filenames, bootleg_annotator)

    def load(self):
        log_model_size(logger, self.server.model, self.server.args.model)
        self.server.model.to(self.server.device)
        self.server.model.eval()
        self.ready = True

    def predict(self, request):
        results = self.server.handle_request(request)
        return {"predictions": results}


def main(args):
    model, device, confidence_estimators, estimator_filenames, bootleg_annotator = init(args)
    model_server = KFModelServer(
        args.inference_name,
        args,
        model.numericalizer,
        model,
        device,
        confidence_estimators,
        estimator_filenames,
        bootleg_annotator,
    )
    model_server.load()
    kfserving.KFServer(workers=1).start([model_server])



import logging
import os
import pickle
import random

import torch
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import Dataset

from ..data_utils.almond_utils import detokenize_cjk_chars
from ..data_utils.progbar import progress_bar
from ..util import get_number_of_lines

logger = logging.getLogger(__name__)


class TextDataset(Dataset):
    def __init__(self, tokenizer, args, file_path=None, block_size=512, evaluate=None):
        self.tokenizer = tokenizer
        self.block_size = block_size
        assert os.path.isfile(file_path)
        directory, filename = os.path.split(file_path)
        cached_features_file = os.path.join(
            directory,
            os.path.basename(os.path.normpath(args.model_name_or_path))
            + '_cached_lm_'
            + str(self.block_size)
            + '_'
            + filename,
        )

        if os.path.exists(cached_features_file) and not args.overwrite_cache:
            logger.info("Loading features from cached file %s", cached_features_file)
            with open(cached_features_file, 'rb') as handle:
                self.input_ids, self.labels, self.position_ids, self.segment_ids = pickle.load(handle)
        else:
            logger.info("Creating features from dataset file at %s", file_path)

            self.prompt_token_id = self.tokenizer.convert_tokens_to_ids(args.start_special_token)
            self.eos_token_id = self.tokenizer.convert_tokens_to_ids(args.end_special_token)
            self.segment1_id = 0
            self.segment2_id = 1
            if args.model_type == 'gpt2':
                self.segment1_id = self.prompt_token_id
                self.segment2_id = self.eos_token_id
            self.input_ids = []
            self.attention_mask = []
            self.labels = []
            self.position_ids = []
            self.segment_ids = []
            self.max_input_length = 0
            self.max_output_length = 0
            self.evaluate = evaluate

            if not self.evaluate and args.aux_train_data_file is not None:
                number_of_lines = get_number_of_lines(args.aux_train_data_file)
                lines = min(args.subsample, number_of_lines)
                i = 0
                with open(args.aux_train_data_file, encoding="utf-8") as f_in:
                    for line in progress_bar(f_in, desc='Tokenizing Auxiliary File', total=number_of_lines):
                        parts = list(map(lambda part: part.strip(), line.split('\t')))
                        i += 1
                        if i > lines:
                            break

                        parts[args.input_column] = args.model_input_prefix + parts[args.input_column]
                        if args.model_type in ['bart', 'mbart', 'marian']:
                            self._add_seq2seq_example(parts[args.input_column], None, args)
                        else:
                            self._add_example(parts[args.input_column], None, args)

            number_of_lines = get_number_of_lines(file_path)
            lines = min(args.subsample, number_of_lines)
            i = 0
            with open(file_path, encoding="utf-8") as f_in:
                for line in progress_bar(f_in, desc='Tokenizing', total=number_of_lines):
                    parts = list(map(lambda part: part.strip(), line.split('\t')))
                    i += 1
                    if i > lines:
                        break

                    parts[args.input_column] = args.model_input_prefix + parts[args.input_column]
                    if args.model_type in ['bart', 'mbart', 'marian']:
                        self._add_seq2seq_example(parts[args.input_column], parts[args.gold_column], args)
                    else:
                        self._add_example(parts[args.input_column], parts[args.gold_column], args)
            if args.sort_by_length:
                _, self.input_ids, self.attention_mask, self.labels, self.position_ids, self.segment_ids = tuple(
                    zip(
                        *sorted(
                            list(
                                zip(
                                    [len(x) for x in self.input_ids],
                                    self.input_ids,
                                    self.attention_mask,
                                    self.labels,
                                    self.position_ids,
                                    self.segment_ids,
                                )
                            )
                        )
                    )
                )
            logger.info('Maximum input length: %d', self.max_input_length)

            if args.cache_input_data:
                logger.info("Saving features into cached file %s", cached_features_file)
                with open(cached_features_file, 'wb') as handle:
                    pickle.dump(
                        (self.input_ids, self.labels, self.position_ids, self.segment_ids),
                        handle,
                        protocol=pickle.HIGHEST_PROTOCOL,
                    )

    def _add_example(self, input_sequence, output_sequence, args):
        # TODO we should make use of tokenizer.build_inputs_with_special_tokens(sequence1, sequence2). Add special tokens manualy only if our model does not support two sequences (like GPT2).

        input_sequence = detokenize_cjk_chars(input_sequence)
        if output_sequence is not None:
            output_sequence = detokenize_cjk_chars(output_sequence)

        input_token_ids = self.tokenizer.encode(input_sequence, add_special_tokens=False) + [
            self.tokenizer.convert_tokens_to_ids(args.start_special_token)
        ]
        if output_sequence is None:
            output_token_ids = []
        else:
            output_token_ids = self.tokenizer.encode(output_sequence, add_special_tokens=False) + [
                self.tokenizer.convert_tokens_to_ids(args.end_special_token)
            ]
        tokenized_text = input_token_ids + output_token_ids

        # do not use exampels that are too long for the model (for supervised tasks, this is better than truncating examples)
        if len(tokenized_text) > self.block_size:
            logger.warning(
                'Skipping example with length %d which was longer than block size (%d)', len(tokenized_text), self.block_size
            )
            return

        input_ids = self.tokenizer.build_inputs_with_special_tokens(tokenized_text)
        # Remove duplicate end_token for models like BERT and RoBERTa that already add it
        if input_ids[-2] == self.eos_token_id:
            input_ids = input_ids[:-1]
        self.max_input_length = max(self.max_input_length, len(input_ids))
        try:
            prompt_token_location = input_ids.index(self.prompt_token_id)
        except ValueError:
            logger.warning('Prompt token not found after truncating the input. Dropping the example.')
            return

        self.input_ids.append(input_ids)
        if args.train_all_tokens and not self.evaluate or output_sequence is None:
            self.labels.append(input_ids)
        else:  # During evaluation, we only care about the output_sequence so we mask the input
            self.labels.append([args.mlm_ignore_index] * (prompt_token_location + 1) + input_ids[prompt_token_location + 1 :])

        position_ids2 = range(len(input_ids) - prompt_token_location - 1)
        if args.reverse_position_ids:
            position_ids2 = reversed(position_ids2)
        self.position_ids.append(list(range(prompt_token_location + 1)) + list(position_ids2))
        self.segment_ids.append(
            [self.segment1_id] * (prompt_token_location + 1)
            + [self.segment2_id] * (len(input_ids) - prompt_token_location - 1)
        )

        # ignored
        self.attention_mask.append([1] * len(input_ids))

    def _add_seq2seq_example(self, input_sequence, output_sequence, args):

        input_sequence = detokenize_cjk_chars(input_sequence)
        if output_sequence is not None:
            output_sequence = detokenize_cjk_chars(output_sequence)

        if args.model_type == 'mbart':
            model_inputs = self.tokenizer.prepare_seq2seq_batch(
                [input_sequence], args.src_lang, [output_sequence], args.tgt_lang
            )
        else:
            model_inputs = self.tokenizer.prepare_seq2seq_batch([input_sequence], [output_sequence], return_tensors='pt')

        encoded_input_ids = model_inputs['input_ids'].tolist()[0]
        encoded_attention_mask = model_inputs['attention_mask'].tolist()[0]
        encoded_output_ids = model_inputs['labels'].tolist()[0]

        self.max_input_length = max(self.max_input_length, len(encoded_input_ids))
        self.max_output_length = max(self.max_output_length, len(encoded_output_ids))

        self.input_ids.append(encoded_input_ids)
        self.attention_mask.append(encoded_attention_mask)
        self.position_ids.append(list(range(len(encoded_input_ids))))
        self.segment_ids.append([self.segment1_id] * len(encoded_input_ids))

        self.labels.append(encoded_output_ids)

    def __len__(self):
        return len(self.input_ids)

    def __getitem__(self, item):
        return (
            torch.tensor(self.input_ids[item]),
            torch.tensor(self.attention_mask[item]),
            torch.tensor(self.labels[item]),
            torch.tensor(self.position_ids[item]),
            torch.tensor(self.segment_ids[item]),
        )

    def collate_fn(self, batch):
        (inputs, attention_mask, labels, position_ids, segment_ids) = zip(*batch)
        inputs_pad = pad_sequence(inputs, batch_first=True, padding_value=self.tokenizer.pad_token_id)
        labels_pad = pad_sequence(labels, batch_first=True, padding_value=-100)
        attention_mask = pad_sequence(attention_mask, batch_first=True, padding_value=0)
        position_ids = pad_sequence(
            position_ids, batch_first=True, padding_value=0
        )  # will be ignored in the loss function, so its value does not matter
        segment_ids = pad_sequence(
            segment_ids, batch_first=True, padding_value=0
        )  # will be ignored in the loss function, so its value does not matter

        return inputs_pad, attention_mask, labels_pad, position_ids, segment_ids


class LengthSortedSampler(torch.utils.data.Sampler):
    def __init__(self, data_source, batch_size, shuffle):
        self.data_source = data_source
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.total_returned_items = 0
        self.last_batch_start_index = 0
        self.last_batch_start_index = self._get_next_batch_start_index()
        self.last_batch_size = 0

    def __len__(self):
        return len(self.data_source)

    def __iter__(self):
        self.total_returned_items = 0
        self.last_batch_start_index = 0
        self.last_batch_start_index = self._get_next_batch_start_index()
        self.last_batch_size = 0
        return self

    def __next__(self):
        ret = (self.last_batch_start_index + self.last_batch_size) % len(self)
        self.total_returned_items = self.total_returned_items + 1
        if self.total_returned_items > len(self):
            raise StopIteration
        self.last_batch_size = self.last_batch_size + 1
        if self.last_batch_size == self.batch_size:
            self.last_batch_size = 0
            self.last_batch_start_index = self._get_next_batch_start_index()

        return ret

    def _get_next_batch_start_index(self):
        if self.shuffle:
            return random.randint(0, len(self))
        else:
            return self.last_batch_start_index + self.batch_size



#
# Copyright (c) 2019-2020 The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


class DecoderVocabulary(object):
    def __init__(self, words, full_vocab, pad_token, eos_token):
        self.full_vocab = full_vocab
        self.pad_token = pad_token
        self.eos_token = eos_token
        stoi = {word: idx for idx, (word, full_idx) in enumerate(words)}
        self.limited_to_full = {stoi[word]: full_idx for word, full_idx in words}
        self.full_to_limited = {full_idx: stoi[word] for word, full_idx in words}
        self.pad_idx = stoi[pad_token]
        self.eos_idx = stoi[eos_token]

    def __len__(self):
        return len(self.limited_to_full)

    def encode(self, full_idx_list):
        limited_list = []
        for full_idx in full_idx_list:
            if full_idx in self.full_to_limited:
                lim_idx = self.full_to_limited[full_idx]
            else:
                lim_idx = len(self)
                self.limited_to_full[lim_idx] = full_idx
                self.full_to_limited[full_idx] = lim_idx
            limited_list.append(lim_idx)
        return limited_list

    def decode(self, lim_idx):
        return self.limited_to_full[lim_idx]



#
# Copyright (c) 2020 The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import torch
from torch import nn

from .common import LayerNorm, LinearFeedforward


class IdentityEncoder(nn.Module):
    def __init__(self, numericalizer, args, config, context_embeddings):
        super().__init__()
        self.args = args
        self.pad_idx = numericalizer.pad_id

        if args.rnn_dimension is None:
            args.rnn_dimension = config.hidden_size

        self.encoder_embeddings = context_embeddings

        if self.args.rnn_layers > 0 and self.args.rnn_dimension != config.hidden_size:
            self.dropout = nn.Dropout(args.dropout_ratio)
            self.projection = nn.Linear(config.hidden_size, self.args.rnn_dimension, bias=False)
        else:
            self.dropout = None
            self.projection = None

        if self.args.rnn_layers > 0 and self.args.rnn_zero_state in ['average', 'cls']:
            self.pool = LinearFeedforward(
                config.hidden_size, config.hidden_size, 2 * args.rnn_dimension * args.rnn_layers, dropout=args.dropout_ratio
            )
            self.norm = LayerNorm(2 * args.rnn_dimension * args.rnn_layers)
        else:
            self.pool = None
            self.norm = None

    def compute_final_embeddings(
        self,
        context,
        context_lengths,
        context_padding,
        context_entity_ids,
        context_entity_probs=None,
        context_entity_masking=None,
        entity_word_embeds_dropout=True,
    ):

        if self.args.do_ned:
            context_embedded_last_hidden_state = self.encoder_embeddings(
                context,
                entity_ids=context_entity_ids,
                entity_masking=context_entity_masking,
                entity_probs=context_entity_probs,
                entity_word_embeds_dropout=entity_word_embeds_dropout,
            ).last_hidden_state

        else:
            context_embedded_last_hidden_state = self.encoder_embeddings(
                context, attention_mask=(~context_padding).to(dtype=torch.float)
            ).last_hidden_state

        final_context = context_embedded_last_hidden_state

        if self.projection is not None:
            final_context = self.dropout(final_context)
            final_context = self.projection(final_context)

        context_rnn_state = None
        if self.args.rnn_layers > 0:
            batch_size = context.size(0)
            if self.args.rnn_zero_state == 'zero':

                zero = torch.zeros(
                    self.args.rnn_layers,
                    batch_size,
                    self.args.rnn_dimension,
                    dtype=torch.float,
                    requires_grad=False,
                    device=context.device,
                )
                context_rnn_state = (zero, zero)
            else:
                if self.args.rnn_zero_state == 'cls':
                    packed_rnn_state = self.norm(self.pool(final_context[:, 0, :]))

                else:
                    assert self.args.rnn_zero_state == 'average'
                    masked_final_context = final_context.masked_fill(context_padding.unsqueeze(2), 0)
                    summed_context = torch.sum(masked_final_context, dim=1)
                    average_context = summed_context / context_lengths.unsqueeze(1)

                    packed_rnn_state = self.norm(self.pool(average_context))

                # packed_rnn_state is (batch, 2 * rnn_layers * rnn_dim)
                packed_rnn_state = packed_rnn_state.reshape(batch_size, 2, self.args.rnn_layers, self.args.rnn_dimension)
                # transpose to (2, batch, rnn_layers, rnn_dimension)
                packed_rnn_state = packed_rnn_state.transpose(0, 1)
                # transpose to (2, rnn_layers, batch, rnn_dimension)
                packed_rnn_state = packed_rnn_state.transpose(1, 2)
                # convert to a tuple of two (rnn_layers, batch, rnn_dimension) tensors
                packed_rnn_state = packed_rnn_state.chunk(2, dim=0)
                context_rnn_state = (packed_rnn_state[0].squeeze(0), packed_rnn_state[1].squeeze(0))

        return final_context, context_rnn_state

    def forward(self, batch):
        context, context_lengths = batch.context.value, batch.context.length
        context_padding = torch.eq(context.data, self.pad_idx)

        context_entity_ids, context_entity_probs, context_entity_masking = None, None, None

        if self.args.num_db_types > 0 and self.args.add_types_to_text == 'no':
            context_entity_ids = batch.context.feature[:, :, : self.args.ned_features_size[0]].long()

            # indicates position of entities
            context_entity_masking = (context_entity_ids != self.args.ned_features_default_val[0]).int()

            if self.args.entity_type_agg_method == 'weighted':
                context_entity_probs = batch.context.feature[
                    :, :, self.args.ned_features_size[0] : self.args.ned_features_size[0] + self.args.ned_features_size[1]
                ].long()

        final_context, context_rnn_state = self.compute_final_embeddings(
            context,
            context_lengths,
            context_padding,
            context_entity_ids,
            context_entity_probs,
            context_entity_masking,
            entity_word_embeds_dropout=self.args.entity_word_embeds_dropout,
        )

        return final_context, context_rnn_state



#
# Copyright (c) 2020 The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
import os

import torch
from transformers import PreTrainedModel

from ..data_utils.numericalizer import TransformerNumericalizer

logger = logging.getLogger(__name__)


class GenieModel(PreTrainedModel):
    numericalizer: TransformerNumericalizer

    @classmethod
    def load(cls, save_directory: str, *model_args, **kwargs):
        """
        Loads a GenieModel (in Genie format, not HuggingFace's transformers) and its
        accompanying Numericalizer (not HuggingFace's tokenizers) from `save_directory`, which is a path
        """
        # TODO remove kwargs and take individual inputs instead
        model_checkpoint_file = kwargs.pop("model_checkpoint_file", None)
        args = kwargs.pop("args", None)
        device = kwargs.pop("device", None)
        tasks = kwargs.pop("tasks", None)
        vocab_sets = kwargs.pop("vocab_sets", None)

        full_checkpoint_path = os.path.join(save_directory, model_checkpoint_file)
        logger.info(f'Loading the model from {full_checkpoint_path}')
        model = cls(args=args, tasks=tasks, vocab_sets=vocab_sets, save_directory=save_directory, *model_args, **kwargs)
        save_dict = torch.load(full_checkpoint_path, map_location=device)

        # HACK
        # `transformers` version 4.1 changed the name of language modeling head of BartForConditionalGeneration
        # (and therefore its subclass MBartForConditionalGeneration) to lm_head to make it similar to other models
        # like T5. The following will make this change so that genienlp models trained with `transformers`==4.0 can be properly loaded
        if (
            'model.lm_head.weight' not in save_dict['model_state_dict']
            and 'model.model.shared.weight' in save_dict['model_state_dict']
        ):
            save_dict['model_state_dict']['model.lm_head.weight'] = save_dict['model_state_dict']['model.model.shared.weight']
        model.load_state_dict(save_dict['model_state_dict'], strict=True)

        return model, save_dict.get('best_decascore')

    def add_new_vocab_from_data(self, tasks, resize_decoder=False):
        old_num_tokens = self.numericalizer.num_tokens
        self.numericalizer.grow_vocab(tasks)
        if self.numericalizer.num_tokens > old_num_tokens:
            logger.info(f'Vocabulary has expanded to {self.numericalizer.num_tokens} tokens')



#
# Copyright (c) 2021, The Board of Trustees of the Leland Stanford Junior University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
import os

import datasets
import torch
from datasets import load_dataset

from ..tasks.generic_dataset import CQA
from .base_dataset import Split

datasets.logging.set_verbosity('ERROR')

logger = logging.getLogger(__name__)


class HFDataset(CQA):
    def __init__(self, data, make_example, **kwargs):

        subsample = kwargs.get('subsample')
        skip_cache = kwargs.pop('kwargs', True)

        cache_name = os.path.join(os.path.dirname(data.cache_files[0]['filename']), data.split._name, str(subsample))
        examples = []

        if os.path.exists(cache_name) and not skip_cache:
            logger.info(f'Loading cached data from {cache_name}')
            examples = torch.load(cache_name)
        for ex in data:
            examples.append(make_example(ex, **kwargs))

            if subsample is not None and len(examples) >= subsample:
                break

        os.makedirs(os.path.dirname(cache_name), exist_ok=True)
        logger.info(f'Caching data to {cache_name}')
        torch.save(examples, cache_name)

        super().__init__(examples, **kwargs)

    @classmethod
    def return_splits(cls, name, root='.data', train='train', validation='validation', test='test', **kwargs):

        # download datasets and cache them
        train_data, validation_data, test_data = None, None, None
        train_path, validation_path, test_path = None, None, None
        if train:
            train_data = load_dataset(name, split='train', cache_dir=root, keep_in_memory=False)
            train_path = train_data.cache_files[0]['filename']
        if validation:
            validation_data = load_dataset(name, split=validation, cache_dir=root, keep_in_memory=False)
            validation_path = validation_data.cache_files[0]['filename']
        if test:
            test_data = load_dataset(name, split='test', cache_dir=root, keep_in_memory=False)
            test_path = test_data.cache_files[0]['filename']

        if kwargs.pop('hf_test_overfit', False):
            # override validation/ test data with train data
            if validation:
                validation_data = load_dataset(name, split='train', cache_dir=root, keep_in_memory=False)
                validation_path = validation_data.cache_files[0]['filename']
            if test:
                test_data = load_dataset(name, split='train', cache_dir=root, keep_in_memory=False)
                test_path = test_data.cache_files[0]['filename']

        train_data = None if train is None else cls(train_data, **kwargs)
        validation_data = None if validation is None else cls(validation_data, **kwargs)
        test_data = None if test is None else cls(test_data, **kwargs)

        return Split(train=train_data, eval=validation_data, test=test_data), Split(
            train=train_path, eval=validation_path, test=test_path
        )
